/* conf.c - Config file reader
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: conf.c,v 1.18.4.2 2001/08/01 12:42:45 entrope Exp $
 */

#include "config.h"
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "recdb.h"
#include "dict.h"
#include "conf.h"
#include "log.h"
#include "common.h"

static dict_t conf_db;
static dict_t conf_server;
static conf_reload_func *reload_funcs;
static int num_rfs, size_rfs;

void
conf_register_reload(conf_reload_func crf)
{
    if (num_rfs >= size_rfs) {
	size_rfs <<= 1;
	reload_funcs = realloc(reload_funcs, size_rfs*sizeof(conf_reload_func));
    }
    reload_funcs[num_rfs++] = crf;
}

void
conf_call_reload_funcs(void)
{
    int i;
    for (i=0; i<num_rfs; i++) reload_funcs[i]();
}

int
conf_read(const char *conf_file_name)
{
    dict_t old_conf = conf_db;
    if (!(conf_db = parse_database(conf_file_name))) {
	goto fail;
    }
    if (!(conf_server = database_get_data(conf_db, "server", RECDB_OBJECT))) {
	log(MAIN_LOG, LOG_ERROR, "No server section found in config file.\n");
	goto fail;
    }
    if (!reload_funcs) {
	size_rfs = 5;
	reload_funcs = malloc(size_rfs*sizeof(conf_reload_func));
	num_rfs = 0;
    } else {
	conf_call_reload_funcs();
    }
    if (old_conf && old_conf != conf_db) free_database(old_conf);
    return 1;

fail:
    log(MAIN_LOG, LOG_ERROR, "Reverting to previous configuration.\n");
    if (conf_db) free_database(conf_db);
    conf_db = old_conf;
    return 0;
}

struct record_data *
conf_get_node(const char *full_path)
{
    return database_get_path(conf_db, full_path);
}

const char*
conf_enum_root(dict_iterator_f it, void *extra)
{
    return dict_foreach(conf_db, it, extra);
}

const char *
conf_server_name(void) {
    return database_get_data(conf_server, "hostname", RECDB_QSTRING);
}

const char *
conf_server_desc(void) {
    return database_get_data(conf_server, "description", RECDB_QSTRING);
}

int
conf_server_num(void) {
    const char *text = database_get_data(conf_server, "numeric", RECDB_QSTRING);
    return text ? atoi(text) : -1;
}

int
conf_server_ping_freq(void) {
    const char *text = database_get_data(conf_server, "ping_freq", RECDB_QSTRING);
    return text ? (signed)ParseInterval(text) : -1;
}

int
conf_server_ping_timeout(void) {
    const char *text = database_get_data(conf_server, "ping_timeout", RECDB_QSTRING);
    return text ? (signed)ParseInterval(text) : -1;
}
