/* dict.c - Abstract dictionary type
 * Copyright 2000 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: dict-splay.c,v 1.1.2.1 2001/06/30 14:39:59 entrope Exp $
 */

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "common.h"
#include "dict.h"

struct dict_node {
    const char *key;
    void *data;
    struct dict_node *l, *r;
};

struct dict {
    free_f free_keys, free_data;
    struct dict_node *root;
    unsigned int count;
};

/*
 *    Create new dictionary.
 */
dict_t
dict_new(void)
{
    dict_t dict;
    dict = malloc(sizeof(struct dict));
    dict->free_keys = NULL;
    dict->free_data = NULL;
    dict->root = NULL;
    dict->count = 0;
    return dict;
}

/*
 *    Return number of entries in the dictionary.
 */
unsigned int
dict_size(dict_t dict)
{
    return dict->count;
}

/*
 *    Set the function to be called when freeing a key structure.
 *    If the function is NULL, just forget about the pointer.
 */
void
dict_set_free_keys(dict_t dict, free_f free_keys)
{
    dict->free_keys = free_keys;
}

/*
 *    Set the function to free data.
 * If the function is NULL, just forget about the pointer.
 */
void
dict_set_free_data(dict_t dict, free_f free_data)
{
    dict->free_data = free_data;
}

/*
 *   This is the daughter function for dict_foreach.
 *   It iterates (in-order) down the node and all its children, calling the
 *   function "it" on each one's key and data (with the "extra" data too).
 *   If the "it" function returns non-zero on any node, return that node
 *   to the caller.
 */
static struct dict_node*
dict_foreach_subtree(struct dict_node *node, dict_iterator_f it, void *extra)
{
    struct dict_node *halter;
    if (!node) return NULL;
    if ((halter = dict_foreach_subtree(node->l, it, extra))) goto out;
    if (it(node->key, node->data, extra)) { halter = node; goto out; }
    halter = dict_foreach_subtree(node->r, it, extra);
 out:
    return halter;
}

/*
 *   Simple wrapper around dict_foreach_subtree, giving a whole-tree
 *   view of the operation.  If "it" returns non-zero for any key/data
 *   pair in the dictionary, return the key on which it did.
 *   (If "it" does ever return non-zero, it will not traverse any more
 *   nodes in the tree.)
 */
const char*
dict_foreach(dict_t dict, dict_iterator_f it, void *extra)
{
    struct dict_node *halter;
    halter = dict_foreach_subtree(dict->root, it, extra);
    return halter ? halter->key : NULL;
}


/*
 *   This function finds a node and pulls it to the top of the tree.
 *   This helps balance the tree and auto-cache things you search for.
 */
static struct dict_node*
dict_splay(struct dict_node *node, const char *key)
{
    struct dict_node N, *l, *r, *y;
    if (!node) return NULL;
    N.l = N.r = NULL;
    l = r = &N;

    while (1) {
	int res = strcasecmp(key, node->key);
	if (!res) break;
	if (res < 0) {
	    if (!node->l) break;
	    res = strcasecmp(key, node->l->key);
	    if (res < 0) {
		y = node->l;
		node->l = y->r;
		y->r = node;
		node = y;
		if (!node->l) break;
	    }
	    r->l = node;
	    r = node;
	    node = node->l;
	} else { /* res > 0 */
	    if (!node->r) break;
	    res = strcasecmp(key, node->r->key);
	    if (res > 0) {
		y = node->r;
		node->r = y->l;
		y->l = node;
		node = y;
		if (!node->r) break;
	    }
	    l->r = node;
	    l = node;
	    node = node->r;
	}
    }
    l->r = node->l;
    r->l = node->r;
    node->l = N.r;
    node->r = N.l;
    return node;
}

/*
 *    Free node.  Free data/key using free_f functions.
 */
static void
dict_dispose_node(struct dict_node *node, free_f free_keys, free_f free_data)
{
    if (free_keys && node->key) free_keys((void*)node->key);
    if (free_data && node->data) free_data(node->data);
    free(node);
}

/*
 *    Insert an entry into the dictionary.
 *    Key ordering (and uniqueness) is determined by case-insensitive
 *    string comparison.
 */
void
dict_insert(dict_t dict, const char *key, void *data)
{
    struct dict_node *new_node;
    if (!key) return;
    new_node = malloc(sizeof(struct dict_node));
    new_node->key = key;
    new_node->data = data;
    if (dict->root) {
	int res;
	dict->root = dict_splay(dict->root, key);
	res = strcasecmp(key, dict->root->key);
	if (res < 0) {
	    new_node->l = dict->root->l;
	    new_node->r = dict->root;
	    dict->root->l = NULL;
	    dict->root = new_node;
	} else if (res > 0) {
	    new_node->r = dict->root->r;
	    new_node->l = dict->root;
	    dict->root->r = NULL;
	    dict->root = new_node;
	} else {
	    /* maybe we don't want to overwrite it .. oh well */
	    if (dict->free_data) dict->free_data(dict->root->data);
            if (dict->free_keys) dict->free_keys((void*)new_node->key);
            free(new_node);
	    dict->root->data = data;
	    /* decrement the count since we dropped the node */
	    dict->count--;
	}
    } else {
	new_node->l = new_node->r = NULL;
	dict->root = new_node;
    }
    dict->count++;
}

/*
 *    Remove an entry from the dictionary.
 *    Return non-zero if it was found, or zero if the key was not in the
 *    dictionary.
 */
int
dict_remove2(dict_t dict, const char *key, int no_dispose)
{
    dict->root = dict_splay(dict->root, key);
    if (dict->root && !strcasecmp(key, dict->root->key)) {
	struct dict_node *new_root;
	if (!dict->root->l) {
	    new_root = dict->root->r;
	} else {
	    new_root = dict_splay(dict->root->l, key);
	    new_root->r = dict->root->r;
	}
        if (no_dispose) {
            free(dict->root);
        } else {
            dict_dispose_node(dict->root, dict->free_keys, dict->free_data);
        }
	dict->root = new_root;
	dict->count--;
	return 1;
    }
    return 0;
}

/*
 *    Find an entry in the dictionary.
 *    If "found" is non-NULL, set it to non-zero if the key was found.
 *    Return the data associated with the key (or NULL if the key was
 *    not found).
 */
void*
dict_find(dict_t dict, const char *key, int *found)
{
    int was_found;
    if (!dict->root || !key) {
	if (found) *found = 0;
	return NULL;
    }
    dict->root = dict_splay(dict->root, key);
    was_found = !strcasecmp(key, dict->root->key);
    if (found) *found = was_found;
    return was_found ? dict->root->data : NULL;
}

/*
 *    Delete a subtree.
 */
static void
delete_subtree(struct dict_node *tree, free_f free_keys, free_f free_data)
{
    if (!tree) return;
    if (tree->l) delete_subtree(tree->l, free_keys, free_data);
    if (tree->r) delete_subtree(tree->r, free_keys, free_data);
    dict_dispose_node(tree, free_keys, free_data);
}

/*
 *    Delete an entire dictionary.
 */
void
dict_delete(dict_t dict)
{
    delete_subtree(dict->root, dict->free_keys, dict->free_data);
    free(dict);
}

struct dict_sanity_struct {
    unsigned int node_count;
    struct dict_node *bad_node;
    char error[128];
};

static int
dict_sanity_check_node(struct dict_node *node, struct dict_sanity_struct *dss)
{
    if (!node->key) {
        snprintf(dss->error, sizeof(dss->error), "Node %p had null key", node);
        return 1;
    }
    if (node->l) {
        if (dict_sanity_check_node(node->l, dss)) return 1;
        if (strcasecmp(node->l->key, node->key) >= 0) {
            snprintf(dss->error, sizeof(dss->error), "Node %p's left child's key '%s' >= its key '%s'", node, node->l->key, node->key);
            return 1;
        }
    }
    if (node->r) {
        if (dict_sanity_check_node(node->r, dss)) return 1;
        if (strcasecmp(node->key, node->r->key) >= 0) {
            snprintf(dss->error, sizeof(dss->error), "Node %p's right child's key '%s' <= its key '%s'", node, node->r->key, node->key);
            return 1;
        }
    }
    dss->node_count++;
    return 0;
}

/*
 *    Perform sanity checks on the dict's internal structure.
 */
char *
dict_sanity_check(dict_t dict)
{
    struct dict_sanity_struct dss;
    dss.node_count = 0;
    dss.bad_node = 0;
    dss.error[0] = 0;
    if (dict->root && dict_sanity_check_node(dict->root, &dss)) {
        return strdup(dss.error);
    } else if (dss.node_count != dict->count) {
        snprintf(dss.error, sizeof(dss.error), "Counted %d nodes but expected %d.", dss.node_count, dict->count);
        return strdup(dss.error);
    } else {
        return 0;
    }
}
